{#
The Mia! Accounting Project
list.html: The unmatched offset list

 Copyright (c) 2023 imacat.

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

Author: imacat@mail.imacat.idv.tw (imacat)
First written: 2023/4/8
#}
{% extends "accounting/base.html" %}

{% block header %}{% block title %}{{ A_("Unmatched Offsets in %(account)s", account=matcher.account.title|title) }}{% endblock %}{% endblock %}

{% block content %}

<div class="btn-group mb-3" role="group" aria-label="{{ A_("Toolbar") }}">
  <a class="btn btn-primary" role="button" href="{{ url_for("accounting.unmatched-offset.dashboard") }}">
    <i class="fa-solid fa-circle-chevron-left"></i>
    {{ A_("Back") }}
  </a>
  {% if matcher.is_having_matches %}
    <button class="btn btn-primary" type="button" data-bs-toggle="modal" data-bs-target="#accounting-match-modal">
      <i class="fa-solid fa-link"></i>
      {{ A_("Match") }}
    </button>
  {% else %}
    <button class="btn btn-secondary" type="button" disabled="disabled">
      <i class="fa-solid fa-link"></i>
      {{ A_("Match") }}
    </button>
  {% endif %}
</div>

{% if matcher.is_having_matches %}
  <form action="{{ url_for("accounting.unmatched-offset.match", account=matcher.account) }}" method="post">
    <input type="hidden" name="csrf_token" value="{{ csrf_token() }}">
    <div class="modal fade" id="accounting-match-modal" tabindex="-1" aria-labelledby="accounting-match-modal-label" aria-hidden="true">
      <div class="modal-dialog">
        <div class="modal-content">
          <div class="modal-header">
            <h1 class="modal-title fs-5" id="accounting-match-modal-label">{{ A_("Confirm Match Offsets") }}</h1>
            <button type="button" class="btn-close" data-bs-dismiss="modal" aria-label="{{ A_("Close") }}"></button>
          </div>
          <div class="modal-body">
            <p>{{ A_("Do you really want to match the following original line items with their offsets?  This cannot be undone.  Please backup your database first, and review before you confirm.") }}</p>

            <ul class="list-group accounting-list-group-stripped accounting-list-group-hover accounting-unmatched-offset-pair-list">
              {% for pair in matcher.matched_pairs %}
                <li class="list-group-item">
                  {{ pair.offset.description|accounting_default }}
                  <span class="badge bg-info">{{ pair.offset.amount|accounting_format_amount }}</span>
                  {{ pair.original_line_item.journal_entry.date|accounting_format_date }} &rarr; {{ pair.offset.journal_entry.date|accounting_format_date }}
                </li>
              {% endfor %}
            </ul>
          </div>
          <div class="modal-footer">
            <button type="button" class="btn btn-secondary" data-bs-dismiss="modal">{{ A_("Cancel") }}</button>
            <button type="submit" class="btn btn-danger">{{ A_("Confirm") }}</button>
          </div>
        </div>
      </div>
    </div>
  </form>
{% endif %}

{% if matcher.total %}
  {% if matcher.is_having_matches %}
    <p>{{ A_("%(matches)s unapplied original line items out of %(total)s can match with their offsets.", matches=matcher.matches, total=matcher.total) }}</p>
  {% else %}
    <p>{{ A_("%(total)s unapplied original line items without matching offsets.", total=matcher.total) }}</p>
  {% endif %}
  <p><a href="{{ url_for("accounting-report.unapplied", account=matcher.account) }}">{{ A_("Go to unapplied original line items.") }}</a></p>
{% else %}
  <p>{{ A_("All original line items are fully offset.") }}</p>
{% endif %}

{% if list %}
  {% include "accounting/include/pagination.html" %}

  <div class="list-group">
  {% for item in list %}
    <a class="list-group-item list-group-item-action {% if not item.match %} list-group-item-danger {% endif %}" href="{{ url_for("accounting.journal-entry.detail", journal_entry=item.journal_entry)|accounting_append_next }}">
      {{ item }}
      {% if item.match %}
        <div class="small">{{ A_("Can match %(item)s", item=item.match) }}</div>
      {% endif %}
    </a>
  {% endfor %}
  </div>
{% else %}
  <p>{{ A_("There is no data.") }}</p>
{% endif %}

{% endblock %}
